/*
 * Copyright 2012 Red Hat, Inc. and/or its affiliates.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
package org.infinispan.arquillian.container.managed;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.concurrent.TimeoutException;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import org.jboss.arquillian.container.spi.client.container.DeployableContainer;
import org.jboss.arquillian.container.spi.client.container.DeploymentException;
import org.jboss.arquillian.container.spi.client.container.LifecycleException;
import org.jboss.arquillian.container.spi.client.protocol.ProtocolDescription;
import org.jboss.arquillian.container.spi.client.protocol.metadata.ProtocolMetaData;
import org.jboss.shrinkwrap.api.Archive;
import org.jboss.shrinkwrap.descriptor.api.Descriptor;

/**
 * A
 * {@link org.jboss.arquillian.container.spi.client.container.DeployableContainer}
 * implementation for the Infinispan container.
 * 
 * {@link InfinispanContainer} is responsible for starting and stopping standalone
 * Infinispan server instances. Each server instance is preconfigured using properties
 * specified in {@link InfinispanConfiguration}. Default values for these properties
 * can be overridden inside Arquillian's configuration file.
 * 
 * @author <a href="mailto:mgencur@redhat.com">Martin Gencur</a>
 * 
 */
public class InfinispanContainer implements DeployableContainer<InfinispanConfiguration>
{
   private static Logger log = Logger.getLogger(InfinispanContainer.class.getName());

   private static final long WAIT_INTERVAL = 200;

   private Process process;

   private InfinispanConfiguration configuration;

   private MBeanServerConnectionProvider provider;

   @Override
   public ProtocolDescription getDefaultProtocol()
   {
      return new ProtocolDescription("Local");
   }

   @Override
   public Class<InfinispanConfiguration> getConfigurationClass()
   {
      return InfinispanConfiguration.class;
   }

   @Override
   public void setup(InfinispanConfiguration configuration)
   {
      this.configuration = configuration;
   }

   @Override
   public void start() throws LifecycleException
   {
      final String ispnHome = configuration.getIspnHome();
      final int jmxPort = configuration.getJmxPort();
      final String mainClassName = "org.infinispan.server.core.Main";
      final String preferIPv4 = "-Djava.net.preferIPv4Stack=true";
      final String log4jConfig = "-Dlog4j.configuration=file://" + ispnHome + File.separator + "etc" + File.separator + "log4j.xml";
      final String jvmJmxSettings = "-Dcom.sun.management.jmxremote.port=" + jmxPort + " -Dcom.sun.management.jmxremote.authenticate=false" + " -Dcom.sun.management.jmxremote.ssl=false";
      try
      {
         List<String> cmd = new ArrayList<String>();

         cmd.add("java");
         cmd.add("-classpath");

         List<File> allLibs = new ArrayList<File>();
         allLibs.addAll(getJars(new File(ispnHome), false));
         allLibs.addAll(getJars(new File(constructFilePath(ispnHome, "lib")), true));

         if ("memcached".equals(configuration.getProtocol()))
         {
            allLibs.addAll(getJars(new File(constructFilePath(ispnHome, "modules", "memcached")), true));
         }
         else if ("hotrod".equals(configuration.getProtocol()))
         {
            allLibs.addAll(getJars(new File(constructFilePath(ispnHome, "modules", "hotrod")), true));
         }
         else if ("websocket".equals(configuration.getProtocol()))
         {
            allLibs.addAll(getJars(new File(constructFilePath(ispnHome, "modules", "websocket")), true));
         }

         cmd.add(constructClassPath(allLibs));

         splitAndAdd(cmd, configuration.getJavaVmArguments());

         cmd.add(preferIPv4);
         cmd.add(log4jConfig);

         splitAndAdd(cmd, jvmJmxSettings);

         cmd.add(mainClassName);

         for (String param : constructProgramArgs())
         {
            cmd.add(param);
         }
         
         splitAndAdd(cmd, configuration.getSystemProperties());

         log.info("Starting container with: " + cmd.toString());
         ProcessBuilder processBuilder = new ProcessBuilder(cmd);
         processBuilder.redirectErrorStream(true);
         process = processBuilder.start();
         new Thread(new ConsoleConsumer()).start();

         provider = new MBeanServerConnectionProvider(InetAddress.getByName(configuration.getHost()), configuration.getJmxPort());
         
         log.info("Waiting for server to start...");
         waitForServerToStart(configuration.getStartupTimeoutInSeconds(), configuration.getProtocol());
      }
      catch (Exception e)
      {
         throw new LifecycleException("Could not start managed container", e);
      }
   }

   private void waitForServerToStart(int startupTimeout, String protocol) throws Exception
   {
      long timeout = startupTimeout * 1000;
      boolean serverAvailable = false;
      while (timeout > 0 && serverAvailable == false)
      {
         if ("hotrod".equals(protocol))
         {
            serverAvailable = isServerModuleStarted("HotRod");
         } 
         else if ("memcached".equals(protocol))
         {
            serverAvailable = isServerModuleStarted("Memcached");
         } 
         else if ("websocket".equals(protocol)) 
         {
            serverAvailable = isDefaultCacheManagerRunning();
         }
         if (!serverAvailable)
         {
            Thread.sleep(WAIT_INTERVAL);
            timeout -= WAIT_INTERVAL;
         }
      }
      if (!serverAvailable)
      {
         destroyProcess();
         throw new TimeoutException(String.format("Managed server was not started within [%d] s", configuration.getStartupTimeoutInSeconds()));
      }
   }

   private boolean isServerModuleStarted(String moduleName) 
   {
       final String serverModule = "org.infinispan:type=Server,name=" + moduleName + ",component=Transport";
       try 
       {
          Set<ObjectInstance> mBeans = (Set<ObjectInstance>) provider.getConnection().queryMBeans(new ObjectName(serverModule), null);
          return mBeans.size() == 1 ? true : false;
       } 
       catch (Exception e) 
       {
          return false;
       }
   }
   
   private boolean isDefaultCacheManagerRunning()
   {
      final String defaultCacheManager = "org.infinispan:type=CacheManager,name=\"DefaultCacheManager\",component=CacheManager";
      final String runningCacheCountAttr = "RunningCacheCount";
      try
      {
         int count = Integer.parseInt(provider.getConnection().getAttribute(new ObjectName(defaultCacheManager), runningCacheCountAttr).toString());
         return count > 0 ? true : false;
      }
      catch (Exception e)
      {
         return false;
      }
   }

   private void splitAndAdd(List<String> cmd, String arg)
   {
      for (String param : arg.split("\\s+"))
      {
         cmd.add(param);
      }
   }

   private List<String> constructProgramArgs()
   {
      List<String> args = new ArrayList<String>();
      args.add("--host=" + configuration.getHost());
      if (configuration.getPort() != 0)
      {
         args.add("--port=" + configuration.getPort());
      }
      if (configuration.getMasterThreads() != 0)
      {
         args.add("--master_threads=" + configuration.getMasterThreads());
      }
      if (configuration.getWorkerThreads() != 0)
      {
         args.add("--worker_threads=" + configuration.getWorkerThreads());
      }
      if (configuration.getCacheConfig() != null)
      {
         args.add("--cache_config=" + configuration.getCacheConfig());
      }
      args.add("--protocol=" + configuration.getProtocol());
      args.add("--idle_timeout=" + configuration.getIdleTimeout());
      args.add("--tcp_no_delay=" + configuration.isTcpNoDelay());
      if (configuration.getSendBufSize() != 0)
      {
         args.add("--send_buf_size=" + configuration.getSendBufSize());
      }
      if (configuration.getRecvBufSize() != 0)
      {
         args.add("--recv_buf_size=" + configuration.getRecvBufSize());
      }
      if (configuration.getProxyHost() != null)
      {
         args.add("--proxy_host=" + configuration.getProxyHost());
      }
      if (configuration.getProxyPort() != 0)
      {
         args.add(" --proxy_port=" + configuration.getProxyPort());
      }
      args.add("--topo_lock_timeout=" + configuration.getTopoLockTimeout());
      args.add("--topo_repl_timeout=" + configuration.getTopoReplTimeout());
      args.add("--topo_state_transfer=" + configuration.isTopoStateTransfer());
      if (configuration.getCacheManagerClass() != null)
      {
         args.add("--cache_manager_class=" + configuration.getCacheManagerClass());
      }
      return args;
   }

   @Override
   public void stop() throws LifecycleException
   {
      try
      {
         if (process != null)
         {
            process.destroy();
            process.waitFor();
            process = null;
         }
      }
      catch (Exception e)
      {
         throw new LifecycleException("Could not stop container", e);
      }
   }

   @Override
   public void deploy(Descriptor descriptor) throws DeploymentException
   {
      throw new UnsupportedOperationException("Deploying to the Infinispan server is not possible");
   }

   @Override
   public void undeploy(Descriptor descriptor) throws DeploymentException
   {
      throw new UnsupportedOperationException("Undeploying from the Infinispan server is not possible");
   }

   @Override
   public ProtocolMetaData deploy(final Archive<?> archive) throws DeploymentException
   {
      throw new UnsupportedOperationException("Deploying to the Infinispan server is not possible");
   }

   @Override
   public void undeploy(final Archive<?> archive) throws DeploymentException
   {
      throw new UnsupportedOperationException("Undeploying from the Infinispan server is not possible");
   }

   private List<File> getJars(File startDir, boolean recursive) throws Exception
   {
      validateDirectory(startDir);
      List<File> result = new ArrayList<File>();

      File[] jars = startDir.listFiles(new FilenameFilter()
      {
         private Pattern jarZipPattern = Pattern.compile(".*\\.([Jj][Aa][Rr]|[Zz][Ii][Pp])$"); // *.jar,*.zip

         public boolean accept(File dir, String name)
         {
            Matcher jarZipMatcher = jarZipPattern.matcher(name);
            return jarZipMatcher.matches();
         }
      });

      List<File> jarList = Arrays.asList(jars);
      result.addAll(jarList);

      if (!recursive)
      {
         return result;
      }

      File[] dirs = startDir.listFiles(new FileFilter()
      {
         public boolean accept(File file)
         {
            return file.isDirectory() && file.canRead();
         }
      });

      for (File file : dirs)
      {
         List<File> deeperList = getJars(file, true);
         result.addAll(deeperList);
      }

      return result;
   }

   private void validateDirectory(File startDir)
   {
      if (!(startDir.isDirectory() && startDir.canRead()))
      {
         throw new IllegalArgumentException("The specified directory does not exist or cannot be read: " + startDir.getAbsolutePath());
      }
   }

   private String constructFilePath(String... strings)
   {
      final String SEP = File.separator;
      StringBuilder sb = new StringBuilder();
      for (String s : strings)
      {
         sb.append(s).append(SEP);
      }
      return sb.substring(0, sb.lastIndexOf(SEP));
   }

   private String constructClassPath(List<File> jarFiles)
   {
      final String SEP = File.pathSeparator;
      StringBuilder sb = new StringBuilder();
      for (File f : jarFiles)
      {
         sb.append(f.getAbsolutePath()).append(SEP);
      }
      return sb.substring(0, sb.lastIndexOf(SEP));
   }

   /**
    * Runnable that consumes the output of the process. If nothing consumes the
    * output the Server will hang on some platforms
    * 
    * @author Stuart Douglas
    */
   private class ConsoleConsumer implements Runnable
   {
      @Override
      public void run()
      {
         final InputStream stream = process.getInputStream();
         final BufferedReader reader = new BufferedReader(new InputStreamReader(stream));
         final boolean writeOutput = AccessController.doPrivileged(new PrivilegedAction<Boolean>()
         {
            @Override
            public Boolean run()
            {
               // By default, redirect to stdout unless disabled by this
               // property
               String val = System.getProperty("org.infinispan.writeconsole");
               return val == null || !"false".equals(val);
            }
         });
         String line = null;
         try
         {
            while ((line = reader.readLine()) != null)
            {
               if (writeOutput)
               {
                  System.out.println(line);
               }
            }
         }
         catch (IOException e)
         {
         }
      }
   }

   private int destroyProcess()
   {
      if (process == null)
      {
         return 0;
      }
      process.destroy();
      try
      {
         return process.waitFor();
      }
      catch (InterruptedException e)
      {
         throw new RuntimeException(e);
      }
   }
}
